<?php
namespace Teknisa\PLA\Modules\Prato\Services;

use Zeedhi\Framework\DTO;
use Zeedhi\Framework\DataSource\DataSet;
use Zeedhi\Framework\DB\StoredProcedure\Param;
use Zeedhi\Framework\DB\StoredProcedure\StoredProcedure;
use PhpOffice\PhpSpreadsheet\IOFactory;

class Pla13000SheetImporter {
    const TMP_DATA_TABLE = "IMP_GRUPOPRATO";
    const SHEETS_FOLDER = __DIR__ . "/../sheets/";
    const SHEET_TEMPLATE = __DIR__ . "/../sheets/Modelo.xlsx";
    protected $COLUMNS_MAP = array(
        "CDFILIAL"      => array(
                                "index"     => 0,
                                "length"    => 4,
                                "nullable"  => false,
                                "zerofill"  => true,
                                "label"     => "Código da Unidade"
                            ),
        "CDSERVFILI"    => array(
                                "index"     => 1,
                                "length"    => 5,
                                "nullable"  => false,
                                "zerofill"  => true,
                                "label"     => "Código do Serviço"
                            ),
        "CDAGRUPAMENTO" => array(
                                "index"     => 2,
                                "length"    => 10,
                                "nullable"  => true,
                                "zerofill"  => true,
                                "label"     => "Código do Agrupamento"
                            ),
        "NMAGRUPAMENTO" => array(
                                "index"     => 3,
                                "maxLength" => 100,
                                "nullable"  => true,
                                "label"     => "Nome do Agrupamento"
                            ),
        "CDGRUPOPRATO"  => array(
                                "index"     => 4,
                                "length"    => 10,
                                "nullable"  => false,
                                "zerofill"  => true,
                                "label"     => "Código do Grupo de Prato"
                            ),
        "NMGRUPOPRATO"  => array(
                                "index"     => 5,
                                "maxLength" => 100,
                                "nullable"  => true,
                                "label"     => "Nome do Grupo de Prato"
                            ),
        "CDPRATO"       => array(
                                "index"     => 6,
                                "maxLength" => 10,
                                "nullable"  => false,
                                "zerofill"  => true,
                                "label"     => "Código do Prato"
                            ),
        "CDFILAUXPRAT"  => array(
                                "index"     => 7,
                                "maxLength" => 4,
                                "nullable"  => false,
                                "zerofill"  => false,
                                "label"     => "Código da Árvore do Prato"
                            )
    );
    protected $errors = array();
    protected $errorMessages = array();
    protected $nrlogimpgrupoprato;

    public function __construct($crudService, $connection, $environment, $dataSourceManager, $restrContrService) {
        $this->crudService = $crudService;
        $this->connection = $connection;
        $this->environment = $environment;
        $this->dataSourceManager = $dataSourceManager;
        $this->restrContrService = $restrContrService;
        $this->COLUMNS_MAP["CDPRATO"]["maxLength"] = $this->getMaxLengthFromMask();
    }

    public function getNrlogimpgrupoprato() {
        return $this->nrlogimpgrupoprato;
    }

    public function downloadTemplate(DTO\Request $request, DTO\Response $response) {
        $name = "Modelo.xlsx";
        header('Accept-Ranges: none');
        header("Content-Type: " . "xlsx");
        header("Content-Disposition: attachment; filename=" . $name);
        readfile(self::SHEET_TEMPLATE);
        exit;
    }

    public function createTemporaryFile($sheetImport) {
        if (    isset($sheetImport) && 
                is_array($sheetImport) && 
                !empty($sheetImport) && 
                isset($sheetImport[0]["b64File"]) && 
                isset($sheetImport[0]["name"])
            ) {
            $sheetB64 = $sheetImport[0]["b64File"];
            $sheetName = $sheetImport[0]["name"];
            $sheet = $this->decodeFromBase64($sheetB64);
            $filePath = self::SHEETS_FOLDER . 'tmp-' . $sheetName;
            file_put_contents($filePath, $sheet);
            chmod($filePath, 0644);
            return $filePath;
        } else {
            throw new \Exception ("sheetUploadFailure");
        }
    }

    public function loadSheet($sheetPath) {
        try {
            $objPHPExcel = IOFactory::load($sheetPath);
            return $objPHPExcel->getSheet(0);
        } catch( \Exception $e) {
            throw new \Exception("invalidSheet");
            
        }
    }

    public function parseSheet($worksheet) {
        try {
            $rows = array();
            foreach ($worksheet->getRowIterator() AS $row) {
                $cellIterator = $row->getCellIterator();
                $cellIterator->resetEnd('H');
                $cellIterator->setIterateOnlyExistingCells(FALSE); // This loops through all cells,
                $cells = [];
                $hasNotNullCell = false;
                foreach ($cellIterator as $cell) {
                    $value = $cell->getValue();
                    $hasNotNullCell = $hasNotNullCell || $value;
                    $cells[] = $value;
                }
                if ($hasNotNullCell) { // At least one cell with value
                    $rows[] = $cells;
                }
            }
            if (count($rows) < 2) {
                throw new \Exception("emptySheet");
            }
            return $rows;
        } catch( \Excention $e) {
            throw $e;
        }
    }

    public function validateSheetColumns($sheet) {
        if (isset($sheet[0])) {
            $header = $sheet[0];
            if (count($header) == 8) {
                foreach ($this->COLUMNS_MAP as $column) {
                    if ($header[$column["index"]] != $column["label"]) {
                        throw new \Exception("invalidSheetColumnsLabels");
                    }
                }
            } else {
                throw new \Exception("invalidSheetColumnsNumber");
            }
        } else {
            throw new \Exception("invalidSheet");
        }
    }

    public function validateSheetRows(&$sheet, $response) {
        foreach ($sheet as $key => &$row) {
            if ($key != 0) { // Skip header row
                $this->validateEmptyValues($row, $key);
                $this->validateInvalidValues($row, $key);
                
            }
        }
        if (!empty($this->errors)) {
            $this->decodeErrorMessages();
            $response->addDataSet(new DataSet("errors", $this->errorMessages));
            throw new \Exception("sheetErrors", 1);
        }
    }

    private function addError($rowNum, $errorType, $error) {
        if (!isset($this->errors[$rowNum])) {
            $this->errors[$rowNum] = array();
        }
        if (!isset($this->errors[$rowNum][$errorType])) {
            $this->errors[$rowNum][$errorType] = array();
        }
        array_push($this->errors[$rowNum][$errorType], $error);
    }

    private function validateEmptyValues(&$row, $rowNum) {
        foreach ($this->COLUMNS_MAP as $columnName => $column) {
            if (is_null($row[$column["index"]]) && !$column["nullable"]) {
                $this->addError($rowNum, "emptyCell", $columnName);
            }
        }
    }

    private function getMaxLengthFromMask() {
        $conditions = array(
            array(
                "columnName" => "NRORG",
                "operator" => "=",
                "value" => $this->environment->getNrOrg()
            )
        );
        $dataset = $this->crudService->getFilteredDataSet("paramgeral", $conditions)->getRows();
        $paramgeral = !empty($dataset) ? $dataset[0] : false;
        $cdpictprat = $paramgeral ? $paramgeral["CDPICTPRAT"] : false;
        if ($cdpictprat) {
            return strlen(str_replace(".", "", $cdpictprat));
        } else {
            return 10;
        }
    }

    private function validateInvalidValues(&$row, $rowNum) {
        foreach ($this->COLUMNS_MAP as $columnName => $column) {
            if (strlen($row[$column["index"]]) != 0) {
                if (isset($column["length"]) && strlen($row[$column["index"]]) != $column["length"]) {
                    if (strlen($row[$column["index"]]) < $column["length"] && isset($column["zerofill"]) && $column["zerofill"]) {
                        $row[$column["index"]] = str_pad($row[$column["index"]], $column["length"], "0", STR_PAD_LEFT);
                    } else {
                        $this->addError($rowNum, "invalidLength", $columnName); 
                    }
                } else if (isset($column["maxLength"]) && strlen($row[$column["index"]]) > $column["maxLength"]) {
                    $this->addError($rowNum, "invalidMaxLength", $columnName); 
                }
            }
        }
    }

    private function decodeErrorMessages() {
        $tab = "          ";
        foreach ($this->errors as $rowNum => $error) {
            $message = "Linha " . ($rowNum + 1) . ":";
            array_push($this->errorMessages, $message);
            if (isset($error["emptyCell"])) {
                foreach (($error["emptyCell"]) as $column) {
                    $message = $tab . "O " . $this->COLUMNS_MAP[$column]["label"] . " deve ser informado.";
                    array_push($this->errorMessages, $message);
                }
            }

            if (isset($error["invalidLength"])) {
                foreach (($error["invalidLength"]) as $column) {
                    $message = $tab . "O " . $this->COLUMNS_MAP[$column]["label"] . " deve possuir exatamente ". $this->COLUMNS_MAP[$column]["length"] . " caracteres.";
                    array_push($this->errorMessages, $message);
                }
            }

            if (isset($error["invalidMaxLength"])) {
                foreach (($error["invalidMaxLength"]) as $column) {
                    $message = $tab . "O " . $this->COLUMNS_MAP[$column]["label"] . " deve possuir no máximo ". $this->COLUMNS_MAP[$column]["maxLength"] . " caracteres.";
                    array_push($this->errorMessages, $message);
                }
            }

            if (isset($error["filialServicoNotFound"])) {
                $message = $tab . "Unidade/Serviço não encontrado.";
                array_push($this->errorMessages, $message);
            }

        }
    }

    private function decodeErrorMessagesFromProcedure($nrlogimpgrupoprato) {
        $conditions = array(
            array(
                "columnName" => "NRLOGIMPGRUPOPRATO",
                "operator"   => "=",
                "value"      => $nrlogimpgrupoprato
            )
        );
        $errors = $this->crudService->getFilteredDataSet("log_imp_grupoprato", $conditions)->getRows();
        usort($errors, function($a, $b) {
			return strnatcmp($a['DSLOGIMPGRUPOPRATO'], $b['DSLOGIMPGRUPOPRATO']);
		});
        foreach ($errors as $error) {
            array_push($this->errorMessages, $error["DSLOGIMPGRUPOPRATO"]);
        }
    }

    private function decodeFromBase64($sheet) {
        $b64File = explode(',', $sheet);
        $base64_decode = base64_decode($b64File[1], true);

        return $base64_decode;
    }

    public function saveSheetData($sheet, $response, $row) {
        $rows = $this->buildRows($sheet, $row);
        return $this->persistData($rows, $response);
    }

    private function buildRows($sheet, $row_param) {
        $nrimpgrupoprato = $this->restrContrService->getNewCode(self::TMP_DATA_TABLE);
        $rows = array();
        $now = $this->environment->getCurrentTime()->format("d/m/Y H:i:s");
        foreach ($sheet as $key => $sheetRow) {
            if ($key != 0) { // Skip header row
                $row = array(
                    "NRIMPGRUPOPRATO" => $nrimpgrupoprato,
                    "NRINDICE"        => $key + 1,
                    "IDTIPORESTRICAO" => $row_param['IDTIPORESTRICAO'],
                    "CDFILIAL"        => $sheetRow[$this->COLUMNS_MAP["CDFILIAL"]["index"]],
                    "CDSERVFILI"      => $sheetRow[$this->COLUMNS_MAP["CDSERVFILI"]["index"]],
                    "CDAGRUPGRPRAT"   => $sheetRow[$this->COLUMNS_MAP["CDAGRUPAMENTO"]["index"]],
                    "NMAGRUPGRPRAT"   => $sheetRow[$this->COLUMNS_MAP["NMAGRUPAMENTO"]["index"]],
                    "CDGRUPOPRATO"    => $sheetRow[$this->COLUMNS_MAP["CDGRUPOPRATO"]["index"]],
                    "NMGRUPOPRATO"    => $sheetRow[$this->COLUMNS_MAP["NMGRUPOPRATO"]["index"]],
                    "CDPRATO"         => $sheetRow[$this->COLUMNS_MAP["CDPRATO"]["index"]],
                    "CDFILAUXPRAT"    => $sheetRow[$this->COLUMNS_MAP["CDFILAUXPRAT"]["index"]],
                    "NRORG"           => $this->environment->getNrOrg(),
                    "IDATIVO"         => "S",
                    "DTINCLUSAO"      => $now,
                    "DTULTATU"        => $now,
                    "NRORGINCLUSAO"   => $this->environment->getNrOrg(),
                    "CDOPERINCLUSAO"  => $this->environment->getCdOperador(),
                    "NRORGULTATU"     => $this->environment->getNrOrg(),
                    "CDOPERULTATU"    => $this->environment->getCdOperador()
                );
                array_push($rows, $row);
            }
        }
        return $rows; 
    }

    private function persistData($rowsToSave, $response) {
        $nrgrupoprato = $rowsToSave[0]["NRIMPGRUPOPRATO"];
        $this->connection->beginTransaction();
        foreach ($rowsToSave as $row) {
            try {
                $this->insertImpgrupoprato($row);
            } catch (\Exception $e) {
                if (strrpos($e->getMessage(), "ORA-02291")) {
                    $this->addError($row["NRINDICE"] - 1, "filialServicoNotFound", true);
                } else {
                    throw $e;
                }
            }
        }
        if (!empty($this->errors)) {
            $this->decodeErrorMessages();
            $response->addDataSet(new DataSet("errors", $this->errorMessages));
            $this->connection->rollBack();
            throw new \Exception("sheetErrors'", 1);
        }
        $this->connection->commit();
        return $nrgrupoprato;
    }

    private function insertImpgrupoprato($row) {
        $sql = "INSERT INTO IMP_GRUPOPRATO 
                    (NRIMPGRUPOPRATO, NRINDICE, IDTIPORESTRICAO, CDFILIAL, CDSERVFILI, CDAGRUPGRPRAT, NMAGRUPGRPRAT, CDGRUPOPRATO, NMGRUPOPRATO, 
                    CDPRATO, CDFILAUXPRAT, NRORG, IDATIVO, DTINCLUSAO, DTULTATU, NRORGINCLUSAO, CDOPERINCLUSAO, NRORGULTATU, CDOPERULTATU)
                VALUES                      
                    (:NRIMPGRUPOPRATO, :NRINDICE, :IDTIPORESTRICAO, :CDFILIAL, :CDSERVFILI, :CDAGRUPGRPRAT, :NMAGRUPGRPRAT, :CDGRUPOPRATO, :NMGRUPOPRATO,
                    :CDPRATO, :CDFILAUXPRAT, :NRORG, :IDATIVO, :DTINCLUSAO, :DTULTATU, :NRORGINCLUSAO, :CDOPERINCLUSAO, :NRORGULTATU, :CDOPERULTATU)
                ";

        $query = $this->connection->prepare($sql);
        $query->bindValue("NRIMPGRUPOPRATO", $row['NRIMPGRUPOPRATO']);
        $query->bindValue("NRINDICE",       $row['NRINDICE']);
        $query->bindValue("IDTIPORESTRICAO",$row['IDTIPORESTRICAO']);
        $query->bindValue("CDFILIAL",       $row['CDFILIAL']);
        $query->bindValue("CDSERVFILI",     $row['CDSERVFILI']);
        $query->bindValue("CDAGRUPGRPRAT",  $row['CDAGRUPGRPRAT']);
        $query->bindValue("NMAGRUPGRPRAT",  $row['NMAGRUPGRPRAT']);
        $query->bindValue("CDGRUPOPRATO",   $row['CDGRUPOPRATO']);
        $query->bindValue("NMGRUPOPRATO",   $row['NMGRUPOPRATO']);
        $query->bindValue("CDPRATO",        $row['CDPRATO']);
        $query->bindValue("CDFILAUXPRAT",   $row['CDFILAUXPRAT']);
        $query->bindValue("NRORG",          $row['NRORG']);
        $query->bindValue("IDATIVO",        $row['IDATIVO']);
        $query->bindValue("DTINCLUSAO",     $row['DTINCLUSAO']);
        $query->bindValue("DTULTATU",       $row['DTULTATU']);
        $query->bindValue("NRORGINCLUSAO",  $row['NRORGINCLUSAO']);
        $query->bindValue("CDOPERINCLUSAO", $row['CDOPERINCLUSAO']);
        $query->bindValue("NRORGULTATU",    $row['NRORGULTATU']);
        $query->bindValue("CDOPERULTATU",   $row['CDOPERULTATU']);
        $query->execute();
    }

    public function executeProcedure($nrimpgrupoprato, $response) {
        $procedure = new StoredProcedure($this->connection, 'IMPORTA_GRUPOPRATO');
        $procedure->addParam(new Param('P_NRIMPGRUPOPRATO', Param::PARAM_INPUT, $nrimpgrupoprato));
        $procedure->addParam(new Param('P_NRORG', Param::PARAM_INPUT, $this->environment->getNrOrg()));
        $procedure->addParam(new Param('P_RETORNO', Param::PARAM_OUTPUT, NULL, NULL, 20));
        $this->nrlogimpgrupoprato = $procedure->execute()["P_RETORNO"];
        
        if (!is_null($this->nrlogimpgrupoprato)) {
            $this->decodeErrorMessagesFromProcedure($this->nrlogimpgrupoprato);
            $response->addDataSet(new DataSet("errors", $this->errorMessages));
            throw new \Exception("procedureErrors", 1);
        }
    }

    public function removeTemporaryFile($filePath) {
        unlink($filePath);
    }

    public function removeSheetTmpData($nrimpgrupoprato) {
        $sql = "DELETE FROM IMP_GRUPOPRATO WHERE NRIMPGRUPOPRATO = :NRIMPGRUPOPRATO AND NRORG = :NRORG";

        $query = $this->connection->prepare($sql);

        $query->bindValue("NRIMPGRUPOPRATO", $nrimpgrupoprato);
        $query->bindValue("NRORG", $this->environment->getNrOrg());
        $query->execute();
    }

    public function removeSheetLog($nrlogimpgrupoprato) {
        $sql = "DELETE FROM LOG_IMP_GRUPOPRATO WHERE NRLOGIMPGRUPOPRATO = :NRLOGIMPGRUPOPRATO AND NRORG = :NRORG";

        $query = $this->connection->prepare($sql);

        $query->bindValue("NRLOGIMPGRUPOPRATO", $nrlogimpgrupoprato);
        $query->bindValue("NRORG", $this->environment->getNrOrg());
        $query->execute();
    }
}
